# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
import unidecode

from majormode.perseus.constant.area import AREA_LEVEL_COUNTRY
from majormode.perseus.constant.area import AREA_LEVEL_WARD
from majormode.perseus.model.label import Label
from majormode.perseus.model.geolocation import BoundingBox
from majormode.perseus.model.geolocation import GeoPoint
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.locale import Locale
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.utils import cast


class AreaService(BaseRdbmsService):
    DEFAULT_DEPTH = 1
    MAXIMUM_DEPTH = 3

    # Define the regular expression that matches a word.
    REGEX_PATTERN_WORD = re.compile('[\W_]+')


    @staticmethod
    def _cleanse_keywords(keywords, keyword_minimal_length=2):
        """
        Remove any punctuation character from the specified list of keywords,
        remove any double or more space character and represent Unicode
        characters in ASCII.


        :param keywords: a list of keywords strip out any punctuation characters.


        :return: the set of keywords cleansed from any special Unicode
            accentuated character, punctuation character, and double space
            character.
        """
        # Normalize the given keywords and split them into sub-keywords if
        # needed.  For instance:
        #
        #   [ u'Saint-Élie-de-Caxton', u'Québec' ]
        #
        # becomes:
        #
        #   [ [ u'saint', u'elie', u'de', u'caxton' ], [ u'Québec' ]]
        sub_keywords_list = [
            re.sub(  # 3. Remove any double space character
                r'\s{2,}', ' ',
                re.sub(  # 2. Remove any punctuation character
                    r"""[.,\\/#!$%\^&*;:{}=\-_`~()<>"']""",
                    ' ',
                    unidecode.unidecode(keyword).lower()))  # 1. Convert to ASCII lowercased characters
            .split(' ')  # 4. Split sub-keywords
            for keyword in keywords]

        # Merge all the sub-keywords in a single list, filtering out
        # sub-keywords of less than 2 characters.
        return set([
            sub_keyword
            for sub_keywords in sub_keywords_list
            for sub_keyword in sub_keywords
            if len(sub_keyword) >= keyword_minimal_length])

    def find_areas_with_ip_address(
            self,
            ip_address,
            connection=None,
            lowest_area_level=5,
            highest_area_level=0,
            locale=DEFAULT_LOCALE):
        """
        Return information of a list of areas that encompass the IP address.


        :param ip_address: a dotted-decimal notation of an IPv4 address,
            consisting of four decimal numbers, each ranging from ``0`` to
            ``255``, separated by dots.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        :param highest_area_level: the level of the largest administrative
            area to finish with.

        :param locale: an instance ``Locale`` representing the locale to
            return the text information of each administrative areas found.


        :return: a list of instances containing the following members:

            * ``area_id`` (required): identification of an administrative
              subdivision.

            * ``parent_area_id`` (optional): identification of the parent
              administrative subdivision, if any.

            * ``area_type`` (required): name of the type of the administrative
              subdivision.  This name can have been localized.  There is no naming
              convention as each country might have its own administrative
              subdivision classification.

            * ``area_level`` (required): administrative level of this area.  For
              clarity and convenience the standard neutral reference for the largest
              administrative subdivision of a country is called the "first-level
              administrative division" or "first administrative level".  Next
              smaller is called "second-level administrative division" or "second
              administrative level".

            * ``label`` (required): an instance ``Label`` of the name of the
              administrative subdivision written in a locale closest to the one
              specified.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    area_id,
                    parent_area_id,
                    ST_X(centroid) AS longitude,
                    ST_Y(centroid) AS latitude,
                    ST_XMin(bounding_box) AS x_min,
                    ST_YMin(bounding_box) AS y_min,
                    ST_XMax(bounding_box) AS x_max,
                    ST_YMax(bounding_box) AS y_max,
                    area_type,
                    area_level,
                    content,
                    get_area_path(area_id, p_locale:=%(locale)s) AS path,
                    locale
                  FROM 
                    find_areas_with_ip_address(
                      %(ip_address)s,
                      p_lowest_area_level:=%(lowest_area_level)s,
                      p_highest_area_level:=%(highest_area_level)s,
                      p_locale:=%(locale)s)
                """,
                {
                    'highest_area_level': highest_area_level,
                    'ip_address': '.'.join([str(byte) for byte in ip_address]),
                    'locale': locale,
                    'lowest_area_level': lowest_area_level
                })

            areas = [
                row.get_object({
                    'area_id': cast.string_to_uuid,
                    'locale': Locale,
                    'parent_area_id': cast.string_to_uuid
                })
                for row in cursor.fetch_all()
            ]

            for area in areas:
                area.label = Label(area.content, area.locale)
                area.path = Label(area.path, area.locale)

                del area.content, area.path, area.locale

                area.centroid = GeoPoint(area.latitude, area.longitude)
                del area.longitude, area.latitude

                area.bounding_box = BoundingBox(
                    GeoPoint(area.y_min, area.x_min),
                    GeoPoint(area.y_max, area.x_max))
                del area.x_min, area.y_min, area.x_max, area.y_max

            return areas

    def find_areas_with_location(
            self,
            location,
            connection=None,
            lowest_area_level=AREA_LEVEL_WARD,
            highest_area_level=AREA_LEVEL_COUNTRY,
            locale=None):
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    area_id,
                    parent_area_id,
                    ST_X(centroid) AS longitude,
                    ST_Y(centroid) AS latitude,
                    ST_XMin(bounding_box) AS x_min,
                    ST_YMin(bounding_box) AS y_min,
                    ST_XMax(bounding_box) AS x_max,
                    ST_YMax(bounding_box) AS y_max,
                    area_type,
                    area_level,
                    content,
                    get_area_path(area_id, p_locale:=%(locale)s) AS path,
                    locale
                  FROM 
                    find_areas_with_location(
                      ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s), 4326),
                      p_lowest_area_level:=%(lowest_area_level)s,
                      p_highest_area_level:=%(highest_area_level)s,
                      p_locale:=%(locale)s)
                """,
                {
                    'highest_area_level': highest_area_level,
                    'latitude': location.latitude,
                    'locale': locale,
                    'longitude': location.longitude,
                    'lowest_area_level': lowest_area_level
                })

            areas = [
                row.get_object({
                    'area_id': cast.string_to_uuid,
                    'locale': Locale,
                    'parent_area_id': cast.string_to_uuid
                })
                for row in cursor.fetch_all()
            ]

            for area in areas:
                area.label = Label(area.content, area.locale)
                area.path = Label(area.path, area.locale)

                del area.content, area.path, area.locale

                area.centroid = GeoPoint(area.latitude, area.longitude)
                del area.longitude, area.latitude

                area.bounding_box = BoundingBox(
                    GeoPoint(area.y_min, area.x_min),
                    GeoPoint(area.y_max, area.x_max))
                del area.x_min, area.y_min, area.x_max, area.y_max

            return areas

    def find_location_with_ip_address(
            self,
            ip_address,
            connection=None):
        """
        Return the location based on the specified IP address.


        :param ip_address: A dotted-decimal notation of an IPv4 address,
            consisting of four decimal numbers, each ranging from ``0`` to
            ``255``, separated by dots.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.


        :return: An instance ``GeoPoint`` corresponding to the location
            associated to the specified IP address.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    ST_X(location) AS longitude,
                    ST_Y(location) AS latitude,
                    ST_Z(location) AS altitude,
                    accuracy
                  FROM
                    geoip_block
                  WHERE
                    network >> %(ip_address)s
                """,
                {
                    'ip_address': '.'.join([str(byte) for byte in ip_address])
                })

            row = cursor.fetch_one()
            location = row and row.get_object()

            return location and GeoPoint(
                location.latitude,
                location.longitude,
                altitude=location.altitude,
                accuracy=location.accuracy)

    def get_area_boundaries(
            self,
            area_id,
            connection=None,
            sync_time=None):
        """
        Return the boundaries of the specified geographic area.


        :param area_id: Identification of the geographic area to return its
            boundaries.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param sync_time: Time of the latest version of the coordinates of the
            boundaries of this geographic area cached by the caller.  If this
            time corresponds to the most recent coordinates of this geographic
            area's boundaries, the request returns `None`, otherwise it
            returns the last version of these coordinates.  If this parameter
            is not provided, the function always returns the most recent
            version of the coordinates of this geographic area's boundaries.


        :return: a list of the coordinates of the boundaries of this
            geographic area:

            * ``[ (lng, lat), ... ]`` for a polygon

            * ``[ [ (lng, lat), ... ], ... ]`` for a multipolygon

            The function returns ``None`` if no geographic area corresponds to
            the specified identification, or if the caller passed the argument
            ``sync_time`` that indicates the caller already cached the most
            recent version of the coordinates of this geographic area's
            boundaries.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    ST_AsText(_boundaries) AS boundaries
                  FROM
                    area
                  WHERE
                    area_id = %(area_id)s
                    AND (%(sync_time)s IS NULL OR %(sync_time)s < update_time)
                """,
                {
                    'area_id': area_id,
                    'sync_time': sync_time
                })
            row = cursor.fetch_one()
            if row is None:
                return None

            boundaries = row.get_value('boundaries')
            if boundaries == 'POLYGON EMPTY' or boundaries == 'MULTIPOLYGON EMPTY':
                return

            # MULTIPOLYGON(((lng lat, ...)), ((lng lat, ...)), ...)
            if boundaries.startswith('MULTIPOLYGON'):
                return [
                    [
                        (float(x), float(y)) for (x, y) in [
                            coordinates.split(' ')
                            for coordinates in polygon.split(',')
                        ]
                    ]
                    for polygon in boundaries[len('MULTIPOLYGON((('):-3].split(')),((')
                ]

            # POLYGON((lng lat, ...))
            elif boundaries.startswith('POLYGON'):
                return [
                    (float(x), float(y)) for (x, y) in [
                        coordinates.split(' ')
                        for coordinates in boundaries[len('POLYGON(('):-2].split(',')
                    ]
                ]

            else:
                raise ValueError(
                    "the boundaries of this area are stored in a wrong format",
                    payload={'area_id': area_id})

    def get_area_extended_id(self, area_id, connection=None):
        """

        :param area_id:

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.


        :return:
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    get_area_extended_id(%(area_id)s) AS area_extended_id
                """,
                {
                    'area_id': area_id
                })
            row = cursor.fetch_one()
            return row and row.get_value('area_extended_id')

    def get_areas(
            self,
            area_ids,
            connection=None,
            locale=DEFAULT_LOCALE):
        """
        Return the specified geographic areas worth of extended information.

        If a requested geographic area is unknown, then that area will not be
        returned in the results list.


        :param area_ids: a list of identifications of geographic areas.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param locale: a ``Locale`` instance that specifies the locale of the
            area labels to be returned.


        :return: a list of instances containing the following members:

            * ``area_id`` (required): identification of the geographic area.

            * ``area_type`` (required): symbolic name of the type of the geographic
              area.

            * ``label`` (required): an instance ``Label`` corresponding to the name
              of this geographic area written in the closest locale to the one
              specified.

            * ``area_level`` (required): administrative level of this geographic
              area.  For clarity and convenience the standard neutral reference for
              the largest administrative subdivision of a country is called the
              "first-level administrative division" or "first administrative level".
              Next smaller is called "second-level administrative division" or
              "second administrative level".

            * ``parent_area_id`` (optional): identification of the parent geographic
              area.
        """
        area_ids = [area_ids] if not isinstance(area_ids, (list, set, tuple)) else list(set(area_ids))

        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    area_id,
                    area_type,
                    area_level,
                    (get_area_label(area_id, %(locale)s)).*,
                    get_area_path(area_id, p_locale:=%(locale)s) AS path,
                    ST_X(centroid) AS longitude,
                    ST_Y(centroid) AS latitude,
                    ST_XMin(bounding_box) AS x_min,
                    ST_YMin(bounding_box) AS y_min,
                    ST_XMax(bounding_box) AS x_max,
                    ST_YMax(bounding_box) AS y_max,
                    parent_area_id
                  FROM
                    area
                  WHERE
                    area_id IN (%(area_ids)s)
                """,
                {
                    'area_ids': area_ids[:self.MAXIMUM_LIMIT],
                    'locale': locale
                })

            areas = [
                row.get_object({
                    'area_id': cast.string_to_uuid,
                    'parent_area_id': cast.string_to_uuid })
                for row in cursor.fetch_all()
            ]

            for area in areas:
                area.label = Label(area.content, area.locale)
                area.path = Label(area.path, area.locale)
                del area.content, area.locale

                area.centroid = GeoPoint(area.latitude, area.longitude)
                del area.longitude, area.latitude

                area.bounding_box = BoundingBox(
                    GeoPoint(area.y_min, area.x_min),
                    GeoPoint(area.y_max, area.x_max))

                del area.x_min, area.y_min, area.x_max, area.y_max

            return areas

    def get_areas_in_bounding_box(
            self,
            bounding_box,
            connection=None,
            locale=DEFAULT_LOCALE):
        """
        Return the geographic areas that intersect the specified bounding box.


        :param bounding_box: a tuple of two instances ``GeoPoint`` that
            represent the north-east corner and the south-west corners of the
            rectangle area to search places in.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param locale: a ``Locale`` instance that specifies the locale of the
             area labels to be returned.


        :return: a list of instances containing the following attributes:

            * ``area_id``: identification of the geographic area.

            * ``area_type``: symbolic name of the type of the geographic area.

            * ``area_label``: label in the specified locale, or the closest locale
              if no label is defined for this particular locale, which is, at
              least, English by default.

            * ``parent_area_id``: identification of the parent geographic area or
              ``None`` if none defined.
        """
        (southwest_location, northeast_location) = bounding_box

        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    area_id,
                    parent_area_id,
                    ST_X(centroid) AS longitude,
                    ST_Y(centroid) AS latitude,
                    ST_XMin(bounding_box) AS x_min,
                    ST_YMin(bounding_box) AS y_min,
                    ST_XMax(bounding_box) AS x_max,
                    ST_YMax(bounding_box) AS y_max,
                    area_type,
                    area_level,
                    content,
                    get_area_path(area_id, p_locale:=%(locale)s) AS path,
                    locale
                  FROM
                    find_areas_in_bounding_box(
                        ST_MakePolygon(ST_GeomFromText(%(bounding_box)s, 4326)),
                        p_locale:=%(locale)s)
                """,
                {
                    'bounding_box': 'LINESTRING({} {},{} {},{} {},{} {},{} {})'.format(
                        southwest_location.longitude, northeast_location.latitude,
                        northeast_location.longitude, northeast_location.latitude,
                        northeast_location.longitude, southwest_location.latitude,
                        southwest_location.longitude, southwest_location.latitude,
                        southwest_location.longitude, northeast_location.latitude),
                    'locale': locale
                })

            areas = [
                row.get_object({
                    'area_id': cast.string_to_uuid,
                    'parent_area_id': cast.string_to_uuid })
                for row in cursor.fetch_all()
            ]

            for area in areas:
                area.label = Label(area.content, area.locale)
                area.path = Label(area.path, area.locale)
                del area.content, area.locale

                area.centroid = GeoPoint(area.latitude, area.longitude)
                del area.longitude, area.latitude

                area.bounding_box = BoundingBox(
                    GeoPoint(area.y_min, area.x_min),
                    GeoPoint(area.y_max, area.x_max))

                del area.x_min, area.y_min, area.x_max, area.y_max

            return areas

    def get_areas_by_ip_address(
            self,
            ip_address,
            connection=None,
            lowest_area_level=AREA_LEVEL_WARD,
            highest_area_level=AREA_LEVEL_COUNTRY,
            minimal_area_surface=None,
            locale=DEFAULT_LOCALE):
        """
        Return the geographic areas that encompass the location determined
        from the specified IP address.


        :param ip_address: a dotted-decimal notation of an IPv4 address,
            consisting of four decimal numbers, each ranging from ``0`` to
            ``255``, separated by dots.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        :param highest_area_level: the level of the largest administrative
            area to finish with.

        :param locale: a ``Locale`` instance that specifies the locale of the
            area labels to be returned.

        :param minimal_area_surface: minimal surface in square meter of the
            geographical area that are returned.


        :return: a list of instances containing the following members:

            * ``area_id`` (required): identification of the geographic area.

            * ``area_label`` (required): an instance ``Label`` corresponding to the
              name of this area in the specified locale, or the closest locale if
              no name is defined for this particular locale, which is, at least,
              English by default.

            * ``area_level`` (required): administrative level of this geogaphic area.
              For clarity and convenience the standard neutral reference for the
              largest administrative subdivision of a country is called the "first-
              level administrative division" or "first administrative level".  Next
              smaller is called "second-level  administrative division" or "second
              administrative level".

            * ``area_type`` (optional): symbolic name of the type of this geographic
              area.

            * ``bounding_box`` (required): an instance ``BoundingBox`` representing
              the maximum extents of the geographical area.

            * ``centroid`` (required): an instance ``GeoPoint`` representing the
              centroid of this area.

            * ``parent_area_id`` (optional): identification of the parent geographic
              area or ``None`` if none defined.
        """
        location = self.find_location_with_ip_address(ip_address, connection=connection)

        return [] if location is None else self.get_areas_by_location(
            location,
            connection=connection,
            lowest_area_level=lowest_area_level,
            highest_area_level=highest_area_level,
            minimal_area_surface=minimal_area_surface,
            locale=locale)

    def get_areas_by_keywords(
            self,
            keywords,
            connection=None,
            highest_area_level=None,
            locale=DEFAULT_LOCALE,
            lowest_area_level=None,
            limit=BaseRdbmsService.DEFAULT_LIMIT,
            minimal_area_surface=None,
            offset=0):
        """
        Return the geographic areas that match the specified keywords.


        :param keywords: a list of one or more keywords.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        :param highest_area_level: The level of the largest administrative
            area to finish with.

        :param limit: Constrain the number of areas that are returned to the
            specified number.

        :param locale: An object ``Locale`` that specifies the locale of the
             area labels to be returned.

        :param minimal_area_surface: Minimal surface of the areas to return.

        :param offset: require to skip that many areas before beginning to
            return them.  If both ``limit`` and ``offset`` are specified, then
            ``offset`` areas are skipped before starting to count the
            `limit`` areas that are returned.  The default value is ``0``.


        :return: an instance containing the following members:

            * ``area_id``: identification of the geographic area.

            * ``area_type``: symbolic name of the type of the geographic area.

            * ``area_label``: label in the specified locale, or the closest locale
              if no label is defined for this particular locale, which is, at
              least, English by default.

            * ``parent_area_id``: identification of the parent geographic area or
              ``None`` if none defined.
        """
        keywords = self._cleanse_keywords(keywords)
        if len(keywords) == 0:
            return []

        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    area_id,
                    parent_area_id,
                    score,
                    ST_X(centroid) AS longitude,
                    ST_Y(centroid) AS latitude,
                    ST_XMin(bounding_box) AS x_min,
                    ST_YMin(bounding_box) AS y_min,
                    ST_XMax(bounding_box) AS x_max,
                    ST_YMax(bounding_box) AS y_max,
                    area_type,
                    area_level,
                    (get_area_label(area_id, %(locale)s)).*,
                    get_area_path(area_id, p_locale:=%(locale)s) AS path
                  FROM (
                    SELECT
                        area_id,
                        COUNT(*) AS score
                      FROM area_index
                      INNER JOIN area
                        USING (area_id)
                      WHERE keyword IN (%[keywords]s)
                        AND (%(highest_area_level)s IS NULL OR area_level >= %(highest_area_level)s)
                        AND (%(lowest_area_level)s IS NULL OR area_level <= %(lowest_area_level)s)
                        AND (%(minimal_area_surface)s IS NULL OR surface >= %(minimal_area_surface)s)
                      GROUP BY area_id
                      ORDER BY
                          score DESC,
                          area_id DESC -- @hack: to preserve order of photos with same score from an offset to another
                      LIMIT %(limit)s
                      OFFSET %(offset)s) AS foo
                  INNER JOIN area
                    USING (area_id)
                """,
                {
                    'keywords': keywords,
                    'highest_area_level': highest_area_level,
                    'limit': min(limit, self.DEFAULT_LIMIT),
                    'locale': locale,
                    'lowest_area_level': lowest_area_level,
                    'minimal_area_surface': minimal_area_surface,
                    'offset': offset
                })

            areas = [
                row.get_object({
                    'area_id': cast.string_to_uuid,
                    'parent_area_id': cast.string_to_uuid,
                    'locale': cast.string_to_locale
                })
                for row in cursor.fetch_all()
            ]

            for area in areas:
                area.label = Label(area.content, area.locale)
                area.path = Label(area.path, area.locale)
                del area.content, area.locale

                area.centroid = GeoPoint(area.latitude, area.longitude)
                del area.longitude, area.latitude

                area.bounding_box = BoundingBox(
                        GeoPoint(area.y_min, area.x_min),
                        GeoPoint(area.y_max, area.x_max))
                del area.x_min, area.y_min, area.x_max, area.y_max

            return areas

    def get_areas_by_location(
            self,
            location,
            connection=None,
            lowest_area_level=AREA_LEVEL_WARD,
            highest_area_level=AREA_LEVEL_COUNTRY,
            locale=None,
            minimal_area_surface=None):
        """
        Return the geographic areas that encompass the specified location.


        :param location: a ``GeoPoint`` instance that specifies a location.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        :param highest_area_level: the level of the largest administrative
            area to finish with.

        :param locale: a ``Locale`` instance that specifies the locale of the
            area labels to be returned.

        :param minimal_area_surface: minimal surface in square meter of the
            geographical area that are returned.


        :return: a list of instances containing the following members:

            * ``area_id`` (required): identification of the geographic area.

            * ``area_label`` (required): an instance ``Label`` corresponding to the
              name of this area in the specified locale, or the closest locale if
              no name is defined for this particular locale, which is, at least,
              English by default.

            * ``area_level`` (required): administrative level of this geogaphic area.
              For clarity and convenience the standard neutral reference for the
              largest administrative subdivision of a country is called the "first-
              level administrative division" or "first administrative level".  Next
              smaller is called "second-level  administrative division" or "second
              administrative level".

            * ``area_type`` (optional): symbolic name of the type of this geographic
              area.

            * ``bounding_box`` (required): an instance ``BoundingBox`` representing
              the maximum extents of the geographical area.

            * ``centroid`` (required): an instance ``GeoPoint`` representing the
              centroid of this area.

            * ``parent_area_id`` (optional): identification of the parent geographic
              area or ``None`` if none defined.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    area_id,
                    parent_area_id,
                    area_level,
                    content,
                    get_area_path(area_id, p_locale:=%(locale)s) AS path,
                    locale,
                    ST_X(centroid) AS longitude,
                    ST_Y(centroid) AS latitude,
                    ST_XMin(bounding_box) AS x_min,
                    ST_YMin(bounding_box) AS y_min,
                    ST_XMax(bounding_box) AS x_max,
                    ST_YMax(bounding_box) AS y_max
                  FROM
                    find_areas_with_location(
                        ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s), 4326),
                        p_lowest_area_level:=%(lowest_area_level)s,
                        p_highest_area_level:=%(highest_area_level)s,
                        p_minimal_area_surface:=%(minimal_area_surface)s,
                        p_locale:=%(locale)s)
                """,
                {
                    'highest_area_level': highest_area_level,
                    'latitude': location.latitude,
                    'locale': locale or DEFAULT_LOCALE,
                    'lowest_area_level': lowest_area_level,
                    'longitude': location.longitude,
                    'minimal_area_surface': minimal_area_surface or 0
                })

            areas = [
                row.get_object({
                    'area_id': cast.string_to_uuid,
                    'locale': Locale,
                    'parent_area_id': cast.string_to_uuid
                })
                for row in cursor.fetch_all()
            ]

            for area in areas:
                area.label = Label(area.content, area.locale)
                area.path = Label(area.path, area.locale)
                del area.content, area.locale

                area.centroid = GeoPoint(area.latitude, area.longitude)
                del area.longitude, area.latitude

                area.bounding_box = BoundingBox(
                    GeoPoint(area.y_min, area.x_min),
                    GeoPoint(area.y_max, area.x_max))

                del area.x_min, area.y_min, area.x_max, area.y_max

            return areas



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.service.account.account_service import AccountService
from majormode.perseus.service.account.test.account_service_unittest import AccountServiceTestCase
from majormode.perseus.service.application.application_service import ApplicationService
from majormode.perseus.unittest.base_unittest import BaseApiKeyTestCase

import hashlib
import hmac
import settings
import unittest
import uuid


class ApplicationServiceTestCase(BaseApiKeyTestCase):
    @staticmethod
    def register_application(account_id):
        """
        Register a new application with a random unique name.

        @param account_id: identification of the account of the user who
               registers this application.

        @return: an instance containing the following members:
                 * ``app_id``: identification of the client application.
                 * ``api_key``:  a unique string that identifies the API key.
                 * ``secret_key``:  a string that must be kept secret, which
                   is used to authenticate the API key.
                 * ``creation_time``: time when the API key has been
                   registered.
       """
        return ApplicationService().register_application(
            account_id,
            'test_%s' % uuid.uuid1().hex,
            ApplicationService.ApplicationPlatform.web)

    """
    Unit testing of the API key service.
    """
    def test_generate_api_key_with_invalid_account(self):
        """
        Generate a random identification of a user's account and try to
        register a new API key, which operation has to raise the exception
        ``UndefinedObjectException``.
        """
        self.assertRaises(AccountService.UndefinedObjectException,
            ApplicationService().register_application,
            uuid.uuid1().hex, 'test', 'mobile')

    def test_generate_api_key_with_duplicate_title(self):
        """
        Register two API keys with the same title, which operation has to
        raise the exception ``ApplicationTitleAlreadyRegisteredException``.
        """
        application_name = uuid.uuid1().hex
        ApplicationService().register_application(self.unittest_account_id, application_name, 'mobile')
        self.assertRaises(ApplicationService.ApplicationNameAlreadyRegisteredException,
            ApplicationService().register_application,
            self.unittest_account_id, application_name, ApplicationService.ApplicationPlatform.mobile)

    def test_assert_invalid_request_signature(self):
        """
        Register an API key and check a virtual API HTTP request providing the
        wrong signature, which operation has to raise the exception
        ``IncorrectSignatureException``.
        """
        application_name = uuid.uuid1().hex
        application = ApplicationService().register_application(self.unittest_account_id, application_name,
            ApplicationService.ApplicationPlatform.mobile)

        api_url = '/resource/01234567890'
        api_sig = hmac.new(str(application.secret_key), msg=api_url, digestmod=hashlib.sha1).hexdigest()
        self.assertRaises(ApplicationService.IncorrectSignatureException,
                          ApplicationService().validate_signature,
                          application.api_key, api_url, None, hashlib.md5(api_sig).hexdigest())

    def test_assert_valid_request_signature(self):
        """
        Register an API key and check a virtual API HTTP request providing the
        correct signature, which operation should succeed.
        """
        application_name = uuid.uuid1().hex
        application = ApplicationService().register_application(self.unittest_account_id, application_name, 'mobile')

        api_url = '/resource/01234567890'
        api_sig = hmac.new(str(application.secret_key), msg=api_url, digestmod=hashlib.sha1).hexdigest()
        ApplicationService().validate_signature(application.api_key, api_url, None, api_sig)

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(ApplicationServiceTestCase))
    return test_suite

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['majormode',
 'majormode.perseus',
 'majormode.perseus.agent',
 'majormode.perseus.bootstrap',
 'majormode.perseus.service',
 'majormode.perseus.service.account',
 'majormode.perseus.service.account.test',
 'majormode.perseus.service.application',
 'majormode.perseus.service.application.test',
 'majormode.perseus.service.area',
 'majormode.perseus.service.area.data.vietnam',
 'majormode.perseus.service.area.script',
 'majormode.perseus.service.notification',
 'majormode.perseus.service.notification.agent',
 'majormode.perseus.service.status',
 'majormode.perseus.service.team',
 'majormode.perseus.service.team.agent',
 'majormode.perseus.service.team.test',
 'majormode.perseus.service.version',
 'majormode.perseus.service.version.test']

package_data = \
{'': ['*'],
 'majormode.perseus': ['db/*'],
 'majormode.perseus.service.account': ['db/*', 'doc/api/*', 'template/*'],
 'majormode.perseus.service.application': ['db/*', 'doc/api/*'],
 'majormode.perseus.service.area': ['db/*', 'doc/api/*'],
 'majormode.perseus.service.notification': ['db/*', 'doc/api/*'],
 'majormode.perseus.service.status': ['doc/api/*'],
 'majormode.perseus.service.team': ['db/*', 'doc/api/*', 'template/*'],
 'majormode.perseus.service.version': ['doc/api/*']}

install_requires = \
['perseus-core-library>=1.18.20,<2.0.0',
 'perseus-getenv-library>=1.0.2,<2.0.0',
 'perseus-microrm-library>=1.3.4,<2.0.0',
 'pillow>=9.3.0,<10.0.0',
 'prosoponym>=1.0.7,<2.0.0',
 'psutil>=5.9.4,<6.0.0',
 'pymemcache>=4.0.0,<5.0.0',
 'tornado>=6.2,<7.0']

setup_kwargs = {
    'name': 'perseus-restful-api-framework',
    'version': '1.25.5',
    'description': 'Python server framework for quickly building RESTful APIs with minimal effort',
    'long_description': '# Perseus: RESTful API Server Framework\n\nPerseus is a Python framework for quickly building RESTful API servers with minimal effort.\n\nPerseus provides an initial set of core services that supports the following features:\n\n- Client application registration with API keys generation\n- Client application access control with RESTful request signature\n- Client application and RESTful API server version compatibility check\n- User authentication and session management\n- Team/group management\n- RESTful request logging with data sensitiveness support\n- RESTful service automatic discovery\n- HTTP request query parameters & body JSON message automatically parsing (depending on the HTTP method used) with data type check and conversion\n\nPerseus is based on [Tornado](https://www.tornadoweb.org/) for handling client network connection.\n\n## RESTful API Request Handler\n\n```python\nfrom majormode.perseus.service.base_http_handler import HttpRequest\nfrom majormode.perseus.service.base_http_handler import HttpRequestHandler\nfrom majormode.perseus.service.base_http_handler import http_request\n\nimport AttendantService\n\n\nclass AttendantServiceHttpRequestHandler(HttpRequestHandler):\n    @http_request(r\'^/attendant/session$\',\n                  http_method=HttpRequest.HttpMethod.POST,\n                  authentication_required=False,\n                  sensitive_data=True,\n                  signature_required=False)\n    def sign_in(self, request):\n        email_address = request.get_argument(\n            \'email_address\',\n            data_type=HttpRequest.ArgumentDataType.email_address,\n            is_required=True)\n\n        password = request.get_argument(\n            \'password\',\n            data_type=HttpRequest.ArgumentDataType.string,\n            is_required=True)\n\n        return AttendantService().sign_in(request.app_id, email_address, password)\n```\n\n## Configure the environment variables\n\n```env\n# Copyright (C) 2021 Majormode.  All rights reserved.\n#\n# Permission is hereby granted, free of charge, to any person obtaining\n# a copy of this software and associated documentation files (the\n# "Software"), to deal in the Software without restriction, including\n# without limitation the rights to use, copy, modify, merge, publish,\n# distribute, sublicense, and/or sell copies of the Software, and to\n# permit persons to whom the Software is furnished to do so, subject to\n# the following conditions:\n#\n# The above copyright notice and this permission notice shall be\n# included in all copies or substantial portions of the Software.\n#\n# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,\n# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF\n# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.\n# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY\n# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,\n# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE\n# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n\n# Connection properties of the RESTful API server instances.  Defaults\n# to 127.0.0.1:8081.\nAPI_SERVER_HOSTNAME=127.0.0.1\nAPI_SERVER_PORTS=\n\n# Root path of the Network File System (NFS) -- referring to the\n# distributed file system (not the protocol) -- where the Content\n# Delivery Network (CDN) files are stored into, such as avatars, etc.\nCDN_NFS_ROOT_PATH=\n\n# Hostname of the Content Delivery Network (CDN) server that hosts media\n# files such as avatars, etc.\nCDN_URL_HOSTNAME=\n\n# Environment stage of the API server instances.  Possible values are:\n#\n# - dev\n# - int\n# - staging\n# - prod\n#\n# Defaults to `dev`.\nENVIRONMENT_STAGE=\n\n# Connection properties to a Memcached server (a distributed memory\n# object caching system).  Defaults to 127.0.0.1:11211.\nMEMCACHED_HOSTNAME = \'127.0.0.1\'\nMEMCACHED_PORT = 11211\n\n# Threshold for the logger to level.  Logging messages which are less\n# severe than the specified level will be ignored; logging messages\n# which have this severity level or higher will be emitted.  Possible\n# values are:\n#\n# - debug\n# - info\n# - warning\n# - error\n# - critical\n#\n# Default to \'debug\'.\nLOGGING_LEVEL=\n\n# Environment variables to select default parameter values to connect\n# to PostgreSQL Relational Database Management System.\nPG_HOSTNAME=localhost\nPG_PORT=5432\nPG_DATABASE_NAME=\nPG_USERNAME=\nPG_PASSWORD=\n```\n\n## Run the RESTful API Server Processes\n\n```bash\n$ fab start --port=65180,65181,...\n```\n\nHashtags/Topics: `#perseus` `#restful` `#api` `#server` `#framework` `#python`\n',
    'author': 'Daniel CAUNE',
    'author_email': 'daniel.caune@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/majormode/perseus-restful-api-server-framework',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.model.version import Version
from majormode.perseus.service.base_service import BaseService

import settings


class VersionService(BaseService):
    """
    This service surfaces the current version of the API deployed on the
    requested environment stage.  This enables a check of the client
    application against the latest version of the API to ensure
    compatibility.  If there is a discrepancy between the two, the client
    application should be prompted to update accordingly.

    The client application verifies compatibility with the latest API
    version through comparing its sequence identifier, which follows the
    form of ``major.minor.patch-build``.  These four variables represent
    the degree of modifications to the API, and will increment based on
    the nature of new developments.
    """
    def get_version(self, app_id):
        """
        Return a ``Version`` instance providing a string representing the
        version of the platform API currently deployed on the environment
        stage that is queried.

        @param app_id: identification of the client application such as a Web,
               a desktop, or a mobile application, that accesses the service.

        @return: a ``Version`` instance.
        """
        return Version(settings.API_VERSION)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with
# the terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE
# SHALL NOT BE LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE
# AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.
#
# @version $Revision$

from majormode.perseus.model import rir
from majormode.utils import cast
from majormode.utils import ftp_util
from majormode.utils import postgresql_util
from majormode.utils import rdbms

import datetime
import dateutil
import os
import settings
import sys


# Minimum time interval between updates, in seconds.
MINIMUM_TIME_INTERVAL_BETWEEN_UPDATES = 60 * 60 * 24 * 30

# Path on the local file system where files downloaded from Regional
# Internet Registry are going to be stored into.
FILE_DOWNLOAD_PATH = '/tmp'

if __name__ == '__main__':
    # Check with Regional Internet Registry has new update to be
    # downloaded.
    updated_rir_statistics = []
    for (rir_name, registry, hostname, remote_path, remote_file_name) in rir.RIR_PROPERTIES:
        with rdbms.RdbmsConnection.acquire_connection(settings.RDBMS_CONNECTION_PROPERTIES) as connection:
            cursor = connection.execute("""
                SELECT md5_checksum,
                       update_time
                  FROM _rir_update
                  WHERE registry = %(registry)s""",
                { 'registry': registry })
            row = cursor.fetch_one()
            last_update = row and row.get_object({ 'update_time': cast.string_to_timestamp })
            if last_update and (datetime.datetime.now(dateutil.tz.tzlocal()) - last_update.update_time).total_seconds() < MINIMUM_TIME_INTERVAL_BETWEEN_UPDATES:
                continue

            sys.stdout.write('Retrieving MD5 checksum for %s...\n' % rir_name)
            md5_file_name ='%s.md5' % remote_file_name
            ftp_util.get_file(hostname, 'anonymous', 'joebar@example.com',
                remote_path, md5_file_name, local_path=FILE_DOWNLOAD_PATH, debug=False)

            with open(os.path.join(FILE_DOWNLOAD_PATH, md5_file_name)) as file:
                md5_checksum = file.read().split('=')[1].strip()

            if last_update is None or md5_checksum != last_update.md5_checksum:
                updated_rir_statistics.append((rir_name, registry, hostname, remote_path, remote_file_name, last_update, md5_checksum))

    # Update the Regional Internet Registry records cached into database.
    if len(updated_rir_statistics) > 0:
        with rdbms.RdbmsConnection.acquire_connection(settings.RDBMS_CONNECTION_PROPERTIES, auto_commit=True) as connection:
            if not postgresql_util.does_table_exist(connection, '_rir_record'):
                connection.execute("""
                    CREATE TABLE _rir_record AS
                      SELECT * FROM rir_record LIMIT 0""") # Trick to create a table with the same structure

            # Remove records that have been lastly downloaded from registries
            # that have new content.
            connection.execute("""
                DELETE FROM _rir_record
                  WHERE registry IN (%[registries]s)""",
                { 'registries': [ registry for (_, registry, _, _, _, _, _) in updated_rir_statistics ] })

            for (rir_name, registry, hostname, remote_path, remote_file_name, last_update, md5_checksum) in updated_rir_statistics:
                sys.stdout.write('Downloading data from %s...' % rir_name)
                ftp_util.get_file(hostname, 'anonymous', 'joebar@example.com',
                    remote_path, remote_file_name, local_path=FILE_DOWNLOAD_PATH, debug=True)
                print

                sys.stdout.write('Processing data from %s...' % rir_name)
                print
                rir_statistics = rir.RIRStatistics.from_file(os.path.join(FILE_DOWNLOAD_PATH, remote_file_name))

                sys.stdout.write('Storing data from %s...' % rir_name)
                print
                connection.execute("""
                    INSERT INTO _rir_record(
                                    registry,
                                    country_code,
                                    record_type,
                                    start,
                                    value,
                                    date,
                                    status)
                      VALUES %[values]s""",
                    { 'values': [ (record.registry, record.country_code, record.type,
                                   record.start, record.value, record.date, record.status)
                          for record in rir_statistics.records ] })

                if last_update:
                    connection.execute("""
                        UPDATE _rir_update
                          SET md5_checksum = %(md5_checksum)s,
                              update_time = current_timestamp
                          WHERE registry = %(registry)s""",
                        { 'md5_checksum': md5_checksum,
                          'registry': registry })
                else:
                    connection.execute("""
                        INSERT INTO _rir_update(
                                        registry,
                                        md5_checksum,
                                        update_time)
                          VALUES (%(registry)s,
                                  %(md5_checksum)s,
                                  current_timestamp)""",
                        { 'md5_checksum': md5_checksum,
                          'registry': registry })

            # Switch a copy of the back table and the front table.
            connection.execute("""
                CREATE TABLE __rir_record AS
                   SELECT * FROM _rir_record""")
            connection.execute('DROP TABLE rir_record')
            connection.execute('ALTER TABLE __rir_record RENAME TO rir_record')



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.service.account.account_service import AccountService
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.utils import cast


class OAuthService(BaseRdbmsService):
    def get_account_with_oauth(self, sns_name, sns_user_id, # sns_app_id
            check_status=False,
            connection=None,
            include_contacts=False):
        """
        Return the account of a user that matches the specified user
        registered on a 3rd party Social Networking Service.


        @param sns_name: code name of the 3rd party Social Networking Service
            such as ``facebook``, ``google``, etc.

        @param sns_user_id: dentification of the user on the 3rd party Social
            Networking Service.

        @param check_status: indicate whether the function MUST check the
            current status of this user account and raise an exception if is
            not of enabled.

        @param connection: a ``RdbmsConnection`` instance to be used
            supporting the Python clause ``with ...:``.

        @param include_contacts: indicate whether to include the contacts
            information of this user account.


        @return: an instance returned by ``AccountService.get_account``.


        @raise DeletedObjectException: if the user account has been deleted,
            while the argument ``check_status`` has been set to ``True``.

        @raise DisabledObjectException: if the user account has been disabled,
            while the argument ``check_status`` has been set to ``True``.

        @raise InvalidOperationException: if this user on the given Social
            Networking Service has linked several accounts on the platform.
            This SHOULD not be possible unless the code of the platform has a
            bug.

        @raise UndefinedObjectException: if the specified email address
            doesn't refer to a user account registered against the platform.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute("""
                SELECT DISTINCT account_id
                  FROM account_oauth
                  WHERE sns_name = %(sns_name)s
                    AND sns_user_id = %(sns_user_id)s
                    AND object_status = %(OBJECT_STATUS_ENABLED)s
                  FROM account_oauth
                  WHERE sns_name = %(sns_name)s
                    AND sns_user_id = %(sns_user_id)s""",
                { 'object_status': ObjectStatus.enabled,
                  'sns_name_': sns_name.strip().lower(),
                  'sns_user_id': sns_user_id })

            row_count = cursor.get_row_count()

            if row_count == 0:
                raise self.UndefinedObjectException('No account registered with the specified Social Networking Service',
                        payload={ 'sns_name': sns_name, 'sns_user_id': sns_user_id })

            if row_count > 1:
                raise self.InvalidOperationException('This SNS user has been linked to multiple accounts',
                        payload={ 'sns_name': sns_name, 'sns_user_id': sns_user_id })

            account_id = row.fetch_one().gat_value('account_id', cast.string_to_uuid)

            return AccountService().get_account(account_id,
                    check_status=check_status,
                    include_contacts=include_contacts,
                    connection=connection)





    def sign_up_with_oauth(self, sns_name, access_token, consumer_key,  consumer_secret,
            auto_sign_in=True,
            locale=None):

        user_info = self.__facebook_get_user_info(access_token)

        if not hasattr(user_info, 'contact'):
            raise self.InvalidOperationException(
                'The platform MUST be given access to the contact information of the user')

        account_session = AccountService().sign_up(
                account_type=AccountType.sns,
                contact=user_info.contact,
                fullname=user_info.name,
                locale=locale,
                auto_sign_in=auto_sign_in,
                has_been_verified=True)




        return account_session



        # # Check whether this user has already linked his Social Networking
        # # Service account with an account registered to the platform.
        # try:
        #     account = OAuthService().get_account_with_oauth(sns_name,
        #                                                     user_info.user_id,
        #                                                     check_status=False,
        #                                                     include_contacts=False)
        #
        # #
        # except OAuthService.UndefinedObjectException:
        #     try:
        #         account = AccountService().get_account_with_contact(
        #             user_info.contact,
        #             check_status=False,
        #             include_pending=True,
        #             is_verified=False)
        #
        #     except AccountService.UndefinedObjectException:








# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.agent.base import BaseAgent
from majormode.perseus.utils import email_util


class EmailSenderAgent(BaseAgent):
    def __init__(self, smtp_connection_properties):
        """
        Build an agent responsible for sending e-mail to an SMTP server.


        :param smtp_connection_properties: An object `SmtpConnection Properties`.
        """
        self.smtp_connection_properties = smtp_connection_properties

    def send_email(
            self,
            email,
            file_path_names=None,
            unsubscribe_mailto_link=None,
            unsubscribe_url=None):
        """
        Send an email to recipients.


        :param email: An object `Mail` electronic mail to be sent to recipient(s).

        :param file_path_names: A list of complete fully qualified path name
            (FQPN) of the files to attach to this message.

        :param unsubscribe_mailto_link: An email address to directly
            unsubscribe the recipient who requests to be removed from the
            mailing list (https://tools.ietf.org/html/rfc2369.html).

            In addition to the email address, other information can be provided.
            In fact, any standard mail header fields can be added to the mailto
            link.  The most commonly used of these are "subject", "cc", and "body"
            (which is not a true header field, but allows you to specify a short
            content message for the new email). Each field and its value is
            specified as a query term (https://tools.ietf.org/html/rfc6068).

        :param unsubscribe_url: A link that will take the subscriber to a
            landing page to process the unsubscribe request.  This can be a
            subscription center, or the subscriber is removed from the list
            right away and gets sent to a landing page that confirms the
            unsubscribe.
        """
        email_util.send_email(
            self.smtp_connection_properties.hostname,
            self.smtp_connection_properties.username,
            self.smtp_connection_properties.password,
            email.author_name,
            email.author_email_address,
            email.recipient_email_addresses,
            email.subject,
            email.content,
            file_path_names=file_path_names,
            port_number=self.smtp_connection_properties.port_number,
            unsubscribe_mailto_link=unsubscribe_mailto_link,
            unsubscribe_url=unsubscribe_url)
